from databricksbundle.notebook.decorator.NodeDecorator import NodeDecorator
from databricksbundle.notebook.decorator.DecoratedDecorator import DecoratedDecorator


@DecoratedDecorator
class data_frame_saver(NodeDecorator):  # noqa: N801
    def __init__(self, *args):
        self._args = args



from logging import Logger
from datalakebundle.hdfs.HdfsExists import HdfsExists
from datalakebundle.table.create.TableCreator import TableCreator
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.delete.TableDeleter import TableDeleter


class TableRecreator:
    def __init__(
        self,
        logger: Logger,
        table_creator: TableCreator,
        table_existence_checker: TableExistenceChecker,
        table_deleter: TableDeleter,
        hdfs_exists: HdfsExists,
    ):
        self.__logger = logger
        self.__table_creator = table_creator
        self.__table_existence_checker = table_existence_checker
        self.__table_deleter = table_deleter
        self.__hdfs_exists = hdfs_exists

    def recreate(self, table_config: TableConfig):
        if self.__table_existence_checker.table_exists(table_config.db_name, table_config.table_name):
            self.__recreate_hive_table(table_config)
        elif self.__hdfs_exists.exists(table_config.target_path):
            self.__create_new_table_when_data_existed(table_config)
        else:
            self.__logger.info(f"Creating new Hive table {table_config.full_table_name} (didn't exist before)")
            self.__table_creator.create_empty_table(table_config)

    def __recreate_hive_table(self, table_config: TableConfig):
        self.__table_deleter.drop_hive_table(table_config)

        if not self.__hdfs_exists.exists(table_config.target_path):
            self.__logger.warning(f"No files in {table_config.target_path} for existing Hive table {table_config.full_table_name}")
        else:
            self.__table_deleter.delete_files(table_config)

        self.__logger.info(f"Recreating Hive table {table_config.full_table_name} (existed before)")
        self.__table_creator.create_empty_table(table_config)

    def __create_new_table_when_data_existed(self, table_config: TableConfig):
        self.__logger.warning(f"Hive table {table_config.full_table_name} does NOT exists for existing files in {table_config.target_path}")

        self.__table_deleter.delete_files(table_config)

        self.__logger.info(f"Creating new Hive table {table_config.full_table_name} (data existed before)")
        self.__table_creator.create_empty_table(table_config)



from datalakebundle.notebook.lineage.DataFrameLoader import DataFrameLoader
from datalakebundle.notebook.lineage.DataFrameSaver import DataFrameSaver
from datalakebundle.notebook.lineage.Transformation import Transformation


def get_mapping():
    return {
        "transformation": Transformation,
        "data_frame_loader": DataFrameLoader,
        "data_frame_saver": DataFrameSaver,
    }



from logging import Logger
from databricksbundle.notebook.decorator.DecoratedDecorator import DecoratedDecorator
from databricksbundle.notebook.decorator.ResultProcessingDecorator import ResultProcessingDecorator
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import DataFrame
from datalakebundle.table.TableManager import TableManager
from datalakebundle.table.TableWriter import TableWriter


@DecoratedDecorator
class table_append(ResultProcessingDecorator):  # noqa: N801
    def __init__(self, table_identifier: str):
        self.__table_identifier = table_identifier

    def process_result(self, result: DataFrame, container: ContainerInterface):
        logger: Logger = container.get("datalakebundle.logger")
        table_manager: TableManager = container.get(TableManager)
        table_writer: TableWriter = container.get(TableWriter)

        output_table_name = table_manager.get_name(self.__table_identifier)

        logger.info(f"Data to be appended into table: {output_table_name}")

        table_manager.create_if_not_exists(self.__table_identifier)

        logger.info(f"Appending data to table: {output_table_name}")

        table_config = table_manager.get_config(self.__table_identifier)

        table_writer.append(result.select([field.name for field in table_config.schema.fields]), table_config)

        logger.info(f"Data successfully appended to: {output_table_name}")



from databricksbundle.notebook.lineage.OutputDecoratorInterface import OutputDecoratorInterface


class TableWriter(OutputDecoratorInterface):
    def __init__(self, table_name: str, mode: str):
        self.__table_name = table_name
        self.__mode = mode

    @property
    def table_name(self):
        return self.__table_name

    @property
    def mode(self):
        return self.__mode



import json
import os
from pathlib import Path
from typing import Union
from injecta.package.path_resolver import resolve_path
from injecta.module import attribute_loader
from datalakebundle.table.schema.DeltaTableSchema import DeltaTableSchema
import pyspark.sql.types as t


class TableConfig:
    def __init__(
        self,
        identifier: str,
        db_identifier: str,
        db_name: str,
        table_identifier: str,
        table_name: str,
        schema_loader: str,
        target_path: str,
        notebook_module: str,
        partition_by: list = None,
        **kwargs,
    ):
        self.__identifier = identifier
        self.__db_identifier = db_identifier
        self.__db_name = db_name
        self.__table_identifier = table_identifier
        self.__table_name = table_name
        self.__schema_loader = schema_loader
        self.__target_path = target_path
        self.__notebook_module = notebook_module
        self.__partition_by = partition_by or []
        self.__custom_fields = kwargs

    @property
    def identifier(self):
        return self.__identifier

    @property
    def db_identifier(self):
        return self.__db_identifier

    @property
    def db_name(self):
        return self.__db_name

    @property
    def table_identifier(self):
        return self.__table_identifier

    @property
    def table_name(self):
        return self.__table_name

    @property
    def schema_loader(self):
        return self.__schema_loader

    @property
    def schema(self) -> Union[t.StructType, DeltaTableSchema]:
        get_schema = attribute_loader.load_from_string(self.__schema_loader)

        return get_schema()

    @property
    def target_path(self):
        return self.__target_path

    @property
    def notebook_module(self):
        return self.__notebook_module

    @property
    def notebook_path(self) -> Path:
        parts = self.__notebook_module.split(".")
        base_path = resolve_path(parts[0])

        return Path(base_path + os.sep + os.sep.join(parts[1:]) + ".py")

    @property
    def partition_by(self):
        return self.__partition_by

    @property
    def full_table_name(self):
        return self.__db_name + "." + self.__table_name

    def get_custom_fields(self):
        return self.__custom_fields

    def as_dict(self):
        return {
            **self.__custom_fields,
            **{
                "identifier": self.__identifier,
                "db_identifier": self.__db_identifier,
                "db_name": self.__db_name,
                "table_identifier": self.__table_identifier,
                "table_name": self.__table_name,
                "schema_loader": self.__schema_loader,
                "target_path": self.__target_path,
                "notebook_module": self.__notebook_module,
                "partition_by": self.__partition_by,
            },
        }

    def __getattr__(self, item):
        if item not in self.__custom_fields:
            raise Exception(f'Unexpected attribute: "{item}"')

        return self.__custom_fields[item]

    def __eq__(self, other: "TableConfig"):
        return self.as_dict() == other.as_dict()

    def __repr__(self):
        return json.dumps(self.as_dict(), indent=4)



import os
from datalakebundle.notebook.decorator.data_frame_loader import data_frame_loader
from datalakebundle.notebook.decorator.transformation import transformation
from datalakebundle.notebook.decorator.data_frame_saver import data_frame_saver

os.environ["APP_ENV"] = "test"


@data_frame_loader()
def load_data():
    return 155


@data_frame_loader()
def load_data2():
    return 145


@transformation(load_data, load_data2, check_duplicate_columns=False)
def sumup(police_number: int, something: int):
    assert police_number == 155
    assert something == 145

    return 155 + 145


@data_frame_saver(sumup)
def save(result: int):
    assert result == 300


if __name__ == "__main__":
    assert isinstance(load_data, data_frame_loader)
    assert load_data.result == 155
    assert globals()["load_data_df"] == 155

    assert isinstance(load_data2, data_frame_loader)
    assert load_data2.result == 145
    assert globals()["load_data2_df"] == 145

    assert isinstance(sumup, transformation)
    assert sumup.result == 300
    assert globals()["sumup_df"] == 300

    assert isinstance(save, data_frame_saver)
    assert save.result is None
    assert "save_df" not in globals()



from databricksbundle.notebook.lineage.argument.DecoratorInputFunctionInterface import DecoratorInputFunctionInterface


class TableParams(DecoratorInputFunctionInterface):
    def __init__(self, table_name: str):
        self.__table_name = table_name

    @property
    def table_name(self):
        return self.__table_name



from databricksbundle.notebook.decorator.notebook_function import notebook_function  # noqa: F401
from databricksbundle.csv.csv_reader import read_csv  # noqa: F401
from datalakebundle.notebook.decorator.data_frame_loader import data_frame_loader  # noqa: F401
from datalakebundle.notebook.decorator.transformation import transformation  # noqa: F401
from datalakebundle.notebook.decorator.data_frame_saver import data_frame_saver  # noqa: F401
from datalakebundle.table.config.table_params import table_params  # noqa: F401
from datalakebundle.table.write.table_append import table_append  # noqa: F401
from datalakebundle.table.write.table_overwrite import table_overwrite  # noqa: F401
from datalakebundle.table.write.table_upsert import table_upsert  # noqa: F401
from datalakebundle.table.read.table_reader import read_table  # noqa: F401



from argparse import Namespace
from logging import Logger
from pyspark.sql import SparkSession
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.config.TableConfigManager import TableConfigManager


class TablesOptimizerCommand(ConsoleCommand):
    def __init__(
        self, logger: Logger, spark: SparkSession, table_config_manager: TableConfigManager, table_existence_checker: TableExistenceChecker
    ):
        self.__logger = logger
        self.__spark = spark
        self.__table_config_manager = table_config_manager
        self.__table_existence_checker = table_existence_checker

    def get_command(self) -> str:
        return "datalake:table:optimize-all"

    def get_description(self):
        return "Runs the OPTIMIZE command on all defined tables (Delta only)"

    def run(self, input_args: Namespace):
        self.__logger.info("Optimizing Hive tables...")

        def filter_func(table_config: TableConfig):
            return self.__table_existence_checker.table_exists(table_config.db_name, table_config.table_name) is True

        existing_tables = self.__table_config_manager.get_by_filter(filter_func)

        self.__logger.info(f"{len(existing_tables)} tables to be optimized")

        for table_config in existing_tables:
            self.__logger.info(f"Running OPTIMIZE {table_config.full_table_name}")

            self.__spark.sql(f"OPTIMIZE {table_config.full_table_name}")



import _ast
from databricksbundle.notebook.lineage.DecoratorParserInterface import DecoratorParserInterface
from datalakebundle.notebook.lineage.TableWriter import TableWriter


class TableWriterParser(DecoratorParserInterface):
    def __init__(self, name: str, mode: str):
        self.__name = name
        self.__mode = mode

    def parse(self, decorator: _ast.Call):
        return TableWriter(decorator.args[0].s, self.__mode)

    def get_name(self) -> str:
        return self.__name



import unittest
from pathlib import Path
from injecta.testing.services_tester import test_services
from pyfonycore.bootstrap import bootstrapped_container
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.config.TableConfigManager import TableConfigManager


class DataLakeBundleTest(unittest.TestCase):
    def setUp(self):
        self._container = bootstrapped_container.init("test")
        self.max_diff = 8000 * 8

    def test_init(self):
        test_services(self._container)

    def test_table_configs(self):
        table_config_manager: TableConfigManager = self._container.get(TableConfigManager)

        my_table_config = table_config_manager.get("mydatabase_e.my_table")
        another_table_config = table_config_manager.get("mydatabase_p.another_table")

        self.assertEqual(
            TableConfig(
                **{
                    "schema_loader": "datalakebundle.test.TestSchema:get_schema",
                    "db_identifier": "mydatabase_e",
                    "table_identifier": "my_table",
                    "identifier": "mydatabase_e.my_table",
                    "notebook_module": "datalakebundle.mydatabase_e.my_table.my_table",
                    "db_name": "test_mydatabase_e",
                    "table_name": "my_table",
                    "encrypted": True,
                    "db_identifier_base": "mydatabase",
                    "target_path": "/foo/bar/mydatabase/encrypted/my_table.delta",
                }
            ),
            my_table_config,
        )

        self.assertEqual(Path.cwd().joinpath("src/datalakebundle/mydatabase_e/my_table/my_table.py"), my_table_config.notebook_path)

        self.assertEqual(
            TableConfig(
                **{
                    "schema_loader": "datalakebundle.test.AnotherSchema:get_schema",
                    "partition_by": ["date"],
                    "db_identifier": "mydatabase_p",
                    "table_identifier": "another_table",
                    "notebook_module": "datalakebundle.mydatabase_p.another_table.another_table",
                    "identifier": "mydatabase_p.another_table",
                    "db_name": "test_mydatabase_p",
                    "table_name": "another_table",
                    "encrypted": False,
                    "db_identifier_base": "mydatabase",
                    "target_path": "/foo/bar/mydatabase/plain/another_table.delta",
                }
            ),
            another_table_config,
        )

        self.assertEqual(
            Path.cwd().joinpath("src/datalakebundle/mydatabase_p/another_table/another_table.py"), another_table_config.notebook_path
        )


if __name__ == "__main__":
    unittest.main()



from pyspark.dbutils import DBUtils


class HdfsExists:
    def __init__(
        self,
        dbutils: DBUtils,
    ):
        self.__dbutils = dbutils

    def exists(self, path: str):
        try:
            self.__dbutils.fs.head(path)

            return True
        except Exception as e:
            if "Cannot head a directory:" in str(e):
                return True

            if "java.io.FileNotFoundException" in str(e):
                return False

            raise



from databricksbundle.notebook.lineage.argument.DecoratorInputFunctionInterface import DecoratorInputFunctionInterface


class ReadCsv(DecoratorInputFunctionInterface):
    def __init__(self, csv_path: str):
        self.__csv_path = csv_path

    @property
    def csv_path(self):
        return self.__csv_path



import sys
from argparse import Namespace


def table_action_command(original_class):
    original_run = original_class.run

    def run(self, input_args: Namespace):
        if not self._table_config_manager.exists(input_args.identifier):
            self._logger.error(f"Identifier {input_args.identifier} not found among datalakebundle.tables")
            sys.exit(1)

        original_run(self, input_args)

    original_class.run = run
    return original_class



from argparse import ArgumentParser, Namespace
from logging import Logger
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.config.TableConfigManager import TableConfigManager
from datalakebundle.table.create.TableRecreator import TableRecreator
from datalakebundle.table.table_action_command import table_action_command


@table_action_command
class TableRecreatorCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        table_config_manager: TableConfigManager,
        table_recreator: TableRecreator,
    ):
        self._logger = logger
        self._table_config_manager = table_config_manager
        self._table_recreator = table_recreator

    def get_command(self) -> str:
        return "datalake:table:recreate"

    def get_description(self):
        return "Recreates a metastore table based on it's YAML definition (name, schema, data path, ...)"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")

    def run(self, input_args: Namespace):
        table_config = self._table_config_manager.get(input_args.identifier)

        self._table_recreator.recreate(table_config)



from pyfonybundles.Bundle import Bundle
from datalakebundle.table.identifier.Expression import Expression

Expression.yaml_loader.add_constructor(Expression.yaml_tag, Expression.from_yaml)


class DataLakeBundle(Bundle):
    pass



from argparse import Namespace
from logging import Logger
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.create.TableCreator import TableCreator
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.config.TableConfigManager import TableConfigManager


class MissingTablesCreatorCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        table_config_manager: TableConfigManager,
        table_creator: TableCreator,
        table_existence_checker: TableExistenceChecker,
    ):
        self.__logger = logger
        self.__table_config_manager = table_config_manager
        self.__table_creator = table_creator
        self.__table_existence_checker = table_existence_checker

    def get_command(self) -> str:
        return "datalake:table:create-missing"

    def get_description(self):
        return "Creates newly defined tables that do not exist in the metastore yet"

    def run(self, input_args: Namespace):
        self.__logger.info("Creating Hive tables...")

        def filter_func(table_config: TableConfig):
            return self.__table_existence_checker.table_exists(table_config.db_name, table_config.table_name) is False

        configs_for_creation = self.__table_config_manager.get_by_filter(filter_func)

        self.__logger.info(f"{len(configs_for_creation)} tables to be created")

        for table_config in configs_for_creation:
            self.__logger.info(f"Creating table {table_config.fullTableName} for {table_config.targetPath}")
            self.__table_creator.create_empty_table(table_config)



from argparse import Namespace, ArgumentParser
from logging import Logger
from pyspark.sql import SparkSession
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.config.TableConfigManager import TableConfigManager
from datalakebundle.table.table_action_command import table_action_command


@table_action_command
class TableOptimizerCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        spark: SparkSession,
        table_config_manager: TableConfigManager,
    ):
        self._logger = logger
        self._spark = spark
        self._table_config_manager = table_config_manager

    def get_command(self) -> str:
        return "datalake:table:optimize"

    def get_description(self):
        return "OPTIMIZEs given table"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")

    def run(self, input_args: Namespace):
        table_config = self._table_config_manager.get(input_args.identifier)

        self._logger.info(f"Running OPTIMIZE {table_config.full_table_name}")

        self._spark.sql(f"OPTIMIZE {table_config.full_table_name}")

        self._logger.info(f"OPTIMIZE for {table_config.full_table_name} completed")



from logging import Logger
from pyspark.dbutils import DBUtils
from pyspark.sql.session import SparkSession
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.UnknownTableException import UnknownTableException
from datalakebundle.table.config.TableConfig import TableConfig


class TableDeleter:
    def __init__(
        self,
        logger: Logger,
        table_existence_checker: TableExistenceChecker,
        spark: SparkSession,
        dbutils: DBUtils,
    ):
        self.__logger = logger
        self.__table_existence_checker = table_existence_checker
        self.__spark = spark
        self.__dbutils = dbutils

    def delete(self, table_config: TableConfig):
        self.__logger.warning(f"HDFS files to be deleted: {table_config.target_path}")

        if self.__table_existence_checker.table_exists(table_config.db_name, table_config.table_name) is False:
            raise UnknownTableException(f"Table {table_config.full_table_name} does not exist in Hive")

        self.drop_hive_table(table_config)
        self.delete_files(table_config)

        self.__logger.info(f"Table {table_config.full_table_name} deleted")

    def drop_hive_table(self, table_config: TableConfig):
        self.__logger.info(f"Deleting Hive table {table_config.full_table_name}")
        self.__spark.sql(f"DROP TABLE {table_config.full_table_name}")
        self.__logger.info(f"Hive table {table_config.full_table_name} deleted")

    def delete_files(self, table_config: TableConfig):
        self.__logger.info(f"Deleting HDFS files from {table_config.target_path}")
        self.__dbutils.fs.rm(table_config.target_path, True)
        self.__logger.info(f"HDFS files deleted from {table_config.target_path}")



from pyspark.sql import SparkSession
from datalakebundle.table.TableWriter import TableWriter
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.hdfs.HdfsExists import HdfsExists


class TableCreator:
    def __init__(
        self,
        spark: SparkSession,
        table_writer: TableWriter,
        table_existence_checker: TableExistenceChecker,
        hdfs_exists: HdfsExists,
    ):
        self.__spark = spark
        self.__table_writer = table_writer
        self.__table_existence_checker = table_existence_checker
        self.__hdfs_exists = hdfs_exists

    def create_empty_table(self, table_config: TableConfig):
        empty_df = self.__spark.createDataFrame([], table_config.schema)

        if self.__table_existence_checker.table_exists(table_config.db_name, table_config.table_name):
            raise Exception(f"Table {table_config.fullTableName} already exists")

        if self.__hdfs_exists.exists(table_config.target_path):
            raise Exception(f"Path {table_config.target_path} already exists")

        self.__table_writer.write_if_not_exist(empty_df, table_config)



import unittest
from pathlib import Path
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.config.TableConfigParser import TableConfigParser


class TableConfigParserTest(unittest.TestCase):
    def setUp(self):
        self.__table_config_parser = TableConfigParser("test_{identifier}")

    def test_basic(self):
        result = self.__table_config_parser.parse(
            "mydatabase.my_table",
            {
                "schema_loader": "datalakebundle.test.TestSchema",
                "target_path": "/foo/bar/mydatabase/my_table.delta",
                "notebook_module": "datalakebundle.foo.mydatabase.my_table.my_table",
                "partition_by": "mycolumn",
            },
        )

        self.assertEqual(
            TableConfig(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "identifier": "mydatabase.my_table",
                    "db_name": "test_mydatabase",
                    "table_name": "my_table",
                    "schema_loader": "datalakebundle.test.TestSchema",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                    "notebook_module": "datalakebundle.foo.mydatabase.my_table.my_table",
                    "partition_by": ["mycolumn"],
                }
            ),
            result,
        )

        self.assertEqual(Path.cwd().joinpath("src/datalakebundle/foo/mydatabase/my_table/my_table.py"), result.notebook_path)

    def test_defaults_only(self):
        result = self.__table_config_parser.parse(
            "mydatabase.my_table",
            {},
            {
                "schema_loader": "datalakebundle.test.{db_identifier}.{table_identifier}.schema:get_schema",
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                },
                "notebook_module": "datalakebundle.foo.{db_identifier}.{table_identifier}.{table_identifier}",
            },
        )

        self.assertEqual(
            TableConfig(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "identifier": "mydatabase.my_table",
                    "db_name": "test_mydatabase",
                    "table_name": "my_table",
                    "schema_loader": "datalakebundle.test.mydatabase.my_table.schema:get_schema",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                    "notebook_module": "datalakebundle.foo.mydatabase.my_table.my_table",
                }
            ),
            result,
        )

        self.assertEqual(Path.cwd().joinpath("src/datalakebundle/foo/mydatabase/my_table/my_table.py"), result.notebook_path)

    def test_explicit_overriding_defaults(self):
        result = self.__table_config_parser.parse(
            "mydatabase.my_table",
            {"schema_loader": "datalakebundle.test.mydatabase.my_table.schema2:get_schema", "partition_by": ["mycolumn"]},
            {
                "schema_loader": "datalakebundle.test.{dbIdentifier}.{tableIdentifier}.schema:get_schema",
                "target_path": {
                    "resolver_class": "datalakebundle.test.SimpleTargetPathResolver",
                    "resolver_arguments": ["/foo/bar"],
                },
                "notebook_module": "datalakebundle.foo.{db_identifier}.{table_identifier}.{table_identifier}",
            },
        )

        self.assertEqual(
            TableConfig(
                **{
                    "db_identifier": "mydatabase",
                    "table_identifier": "my_table",
                    "identifier": "mydatabase.my_table",
                    "db_name": "test_mydatabase",
                    "table_name": "my_table",
                    "schema_loader": "datalakebundle.test.mydatabase.my_table.schema2:get_schema",
                    "target_path": "/foo/bar/mydatabase/my_table.delta",
                    "notebook_module": "datalakebundle.foo.mydatabase.my_table.my_table",
                    "partition_by": ["mycolumn"],
                }
            ),
            result,
        )

        self.assertEqual(Path.cwd().joinpath("src/datalakebundle/foo/mydatabase/my_table/my_table.py"), result.notebook_path)


if __name__ == "__main__":
    unittest.main()



from databricksbundle.notebook.lineage.InputDecoratorInterface import InputDecoratorInterface


class Transformation(InputDecoratorInterface):
    def __init__(self, name, args):
        self.__name = name
        self.__args = args

    @property
    def name(self):
        return self.__name

    @property
    def args(self):
        return self.__args



from pyspark.sql.types import StructType


class DeltaTableSchema(StructType):
    def __init__(self, fields: list, primary_key, constraints: list = None):
        super().__init__(fields)
        self.__primary_key_columns = [primary_key] if isinstance(primary_key, str) else primary_key
        self.__constraints = constraints or []

    @property
    def primary_key_columns(self) -> list:
        return self.__primary_key_columns

    @property
    def constraints(self):
        return self.__constraints

    @classmethod
    def typeName(cls):  # noqa: N802
        return "struct"



from argparse import ArgumentParser, Namespace
from logging import Logger
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.config.TableConfigManager import TableConfigManager
from datalakebundle.table.create.TableCreator import TableCreator
from datalakebundle.table.table_action_command import table_action_command


@table_action_command
class TableCreatorCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        table_config_manager: TableConfigManager,
        table_creator: TableCreator,
    ):
        self._logger = logger
        self._table_config_manager = table_config_manager
        self._table_creator = table_creator

    def get_command(self) -> str:
        return "datalake:table:create"

    def get_description(self):
        return "Creates a metastore table based on it's YAML definition (name, schema, data path, ...)"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")

    def run(self, input_args: Namespace):
        table_config = self._table_config_manager.get(input_args.identifier)

        self._logger.info(f"Creating table {table_config.fullTableName} for {table_config.targetPath}")

        self._table_creator.create_empty_table(table_config)

        self._logger.info(f"Table {table_config.fullTableName} successfully created")



import unittest
from types import FunctionType


class DecoratorsTest(unittest.TestCase):
    def test_basic(self):
        from datalakebundle.notebook.decorator.tests.decorators_test import load_data

        result = load_data()

        self.assertIsInstance(load_data, FunctionType)
        self.assertEqual(155, result)

    def test_error(self):
        with self.assertRaises(Exception) as error:
            from datalakebundle.notebook.decorator.tests.decorators_fixture import load_data3  # noqa: F401

        self.assertEqual("Use @data_frame_loader() instead of @data_frame_loader please", str(error.exception))


if __name__ == "__main__":
    unittest.main()



from logging import Logger
from databricksbundle.notebook.decorator.DecoratedDecorator import DecoratedDecorator
from databricksbundle.notebook.decorator.ResultProcessingDecorator import ResultProcessingDecorator
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import DataFrame
from datalakebundle.table.TableManager import TableManager
from datalakebundle.table.TableWriter import TableWriter


@DecoratedDecorator
class table_overwrite(ResultProcessingDecorator):  # noqa: N801
    def __init__(self, table_identifier: str, recreate_table=False):
        self.__table_identifier = table_identifier
        self.__recreate_table = recreate_table

    def process_result(self, result: DataFrame, container: ContainerInterface):
        logger: Logger = container.get("datalakebundle.logger")
        table_manager: TableManager = container.get(TableManager)
        table_writer: TableWriter = container.get(TableWriter)

        output_table_name = table_manager.get_name(self.__table_identifier)

        logger.info(f"Data to be written into table: {output_table_name}")

        if self.__recreate_table:
            table_manager.recreate(self.__table_identifier)
        else:
            table_manager.create_if_not_exists(self.__table_identifier)

        logger.info(f"Writing data to table: {output_table_name}")

        table_config = table_manager.get_config(self.__table_identifier)

        table_writer.overwrite(result.select([field.name for field in table_config.schema.fields]), table_config)

        logger.info(f"Data successfully written to: {output_table_name}")



from box import Box
from datalakebundle.table.identifier import placeholders_extractor
from datalakebundle.table.identifier.ValueResolverInterface import ValueResolverInterface


class NotebookModuleResolver(ValueResolverInterface):
    def __init__(
        self,
        notebook_module_template: str,
    ):
        self.__notebook_module_template = notebook_module_template

    def resolve(self, raw_table_config: Box):
        return self.__notebook_module_template.format(**raw_table_config)

    def get_depending_fields(self) -> set:
        return placeholders_extractor.extract(self.__notebook_module_template)



from box import Box
from datalakebundle.table.identifier import placeholders_extractor
from datalakebundle.table.identifier.ValueResolverInterface import ValueResolverInterface


class SchemaLoaderResolver(ValueResolverInterface):
    def __init__(
        self,
        schema_loader_template: str,
    ):
        self.__schema_loader_template = schema_loader_template

    def resolve(self, raw_table_config: Box):
        raw_table_config["notebook_parent_module"] = self._get_parent_module(raw_table_config["notebook_module"])

        return self.__schema_loader_template.format(**raw_table_config)

    def get_depending_fields(self) -> set:
        fields = placeholders_extractor.extract(self.__schema_loader_template)
        fields.add("notebook_module")

        if "notebook_parent_module" in fields:
            fields.remove("notebook_parent_module")

        return fields

    def _get_parent_module(self, notebook_module: str):
        return notebook_module[: notebook_module.rfind(".")]



from databricksbundle.notebook.decorator.DecoratedDecorator import DecoratedDecorator
from injecta.container.ContainerInterface import ContainerInterface
from databricksbundle.display import display as display_function
from datalakebundle.notebook.decorator.DataFrameReturningDecorator import DataFrameReturningDecorator


@DecoratedDecorator
class data_frame_loader(DataFrameReturningDecorator):  # noqa: N801
    def __init__(self, *args, display=False):
        self._args = args
        self._display = display

    def after_execution(self, container: ContainerInterface):
        self._set_global_dataframe()

        if self._display and container.get_parameters().datalakebundle.notebook.display.enabled is True:
            display_function(self._result)



from logging import Logger
from injecta.container.ContainerInterface import ContainerInterface
from pyspark.sql import SparkSession
from databricksbundle.notebook.function.node_argument_function import node_argument_function
from datalakebundle.table.TableManager import TableManager


@node_argument_function
def read_table(identifier: str):
    def wrapper(container: ContainerInterface):
        table_manager: TableManager = container.get(TableManager)
        table_name = table_manager.get_name(identifier)

        logger: Logger = container.get("datalakebundle.logger")
        logger.info(f"Reading table `{table_name}`")

        spark: SparkSession = container.get(SparkSession)

        return spark.read.table(table_name)

    return wrapper



from argparse import Namespace
from logging import Logger
from datalakebundle.table.TableExistenceChecker import TableExistenceChecker
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.config.TableConfigManager import TableConfigManager
from datalakebundle.table.create.TableCreator import TableCreator
from datalakebundle.table.create.TableRecreator import TableRecreator
from datalakebundle.table.delete.TableDeleter import TableDeleter
from datalakebundle.table.identifier.fill_template import fill_template
from datalakebundle.table.optimize.TablesOptimizerCommand import TablesOptimizerCommand


class TableManager:
    def __init__(
        self,
        table_name_template: str,
        logger: Logger,
        table_config_manager: TableConfigManager,
        table_creator: TableCreator,
        table_recreator: TableRecreator,
        table_deleter: TableDeleter,
        table_existence_checker: TableExistenceChecker,
        tables_optimizer_command: TablesOptimizerCommand,
    ):
        self.__table_name_template = table_name_template
        self.__logger = logger
        self.__table_config_manager = table_config_manager
        self.__table_creator = table_creator
        self.__table_recreator = table_recreator
        self.__table_deleter = table_deleter
        self.__table_existence_checker = table_existence_checker
        self.__tables_optimizer_command = tables_optimizer_command

    def get_name(self, identifier: str):
        table_config = self.__table_config_manager.get(identifier)

        replacements = {
            **table_config.get_custom_fields(),
            **{
                "identifier": table_config.identifier,
                "db_identifier": table_config.db_identifier,
                "table_identifier": table_config.table_identifier,
            },
        }

        return fill_template(self.__table_name_template, replacements)

    def get_config(self, identifier: str) -> TableConfig:
        return self.__table_config_manager.get(identifier)

    def create(self, identifier: str):
        table_config = self.__table_config_manager.get(identifier)

        self.__create(table_config)

    def create_if_not_exists(self, identifier: str):
        table_config = self.__table_config_manager.get(identifier)

        if self.__table_existence_checker.table_exists(table_config.db_name, table_config.table_name):
            self.__logger.info(f"Table {table_config.full_table_name} already exists, creation skipped")
            return

        self.__create(table_config)

    def recreate(self, identifier: str):
        table_config = self.__table_config_manager.get(identifier)

        self.__table_recreator.recreate(table_config)

    def exists(self, identifier: str):
        table_config = self.__table_config_manager.get(identifier)

        return self.__table_existence_checker.table_exists(table_config.db_name, table_config.table_name)

    def delete(self, identifier: str):
        table_config = self.__table_config_manager.get(identifier)

        self.__table_deleter.delete(table_config)

        self.__logger.info(f"Table {table_config.full_table_name} successfully deleted")

    def optimize_all(self):
        self.__tables_optimizer_command.run(Namespace())

    def __create(self, table_config: TableConfig):
        self.__logger.info(f"Creating table {table_config.full_table_name} for {table_config.target_path}")

        self.__table_creator.create_empty_table(table_config)

        self.__logger.info(f"Table {table_config.full_table_name} successfully created")



from logging import Logger
from databricksbundle.notebook.decorator.DecoratedDecorator import DecoratedDecorator
from databricksbundle.notebook.decorator.ResultProcessingDecorator import ResultProcessingDecorator
from injecta.container.ContainerInterface import ContainerInterface


class TestingStorage:
    result: None


@DecoratedDecorator
class dummy_saver(ResultProcessingDecorator):  # noqa: N801
    def __init__(self, table_identifier: str):
        self._table_identifier = table_identifier

    def process_result(self, result, container: ContainerInterface):
        logger: Logger = container.get("datalakebundle.logger")

        logger.info(f"Saving into {self._table_identifier}")

        TestingStorage.result = result



from databricksbundle.notebook.decorator.DecoratedDecorator import DecoratedDecorator
from injecta.container.ContainerInterface import ContainerInterface
from databricksbundle.display import display as display_function
from datalakebundle.notebook.decorator.DataFrameReturningDecorator import DataFrameReturningDecorator
from datalakebundle.notebook.decorator.DuplicateColumnsChecker import DuplicateColumnsChecker


@DecoratedDecorator
class transformation(DataFrameReturningDecorator):  # noqa: N801
    def __init__(self, *args, display=False, check_duplicate_columns=True):
        self._args = args
        self._display = display
        self._check_duplicate_columns = check_duplicate_columns

    def after_execution(self, container: ContainerInterface):
        if self._check_duplicate_columns and container.get_parameters().datalakebundle.notebook.duplicate_columns_check.enabled is True:
            duplicate_columns_checker: DuplicateColumnsChecker = container.get(DuplicateColumnsChecker)

            data_frame_decorators = tuple(
                decorator_arg for decorator_arg in self._args if isinstance(decorator_arg, DataFrameReturningDecorator)
            )
            duplicate_columns_checker.check(self._result, data_frame_decorators)

        self._set_global_dataframe()

        if self._display and container.get_parameters().datalakebundle.notebook.display.enabled is True:
            display_function(self._result)



import sys
from argparse import ArgumentParser, Namespace
from logging import Logger
from consolebundle.ConsoleCommand import ConsoleCommand
from datalakebundle.table.UnknownTableException import UnknownTableException
from datalakebundle.table.config.TableConfigManager import TableConfigManager
from datalakebundle.table.delete.TableDeleter import TableDeleter
from datalakebundle.table.table_action_command import table_action_command


@table_action_command
class TableDeleterCommand(ConsoleCommand):
    def __init__(
        self,
        logger: Logger,
        table_config_manager: TableConfigManager,
        table_deleter: TableDeleter,
    ):
        self._logger = logger
        self._table_config_manager = table_config_manager
        self._table_deleter = table_deleter

    def get_command(self) -> str:
        return "datalake:table:delete"

    def get_description(self):
        return "Deletes a metastore table including data on HDFS"

    def configure(self, argument_parser: ArgumentParser):
        argument_parser.add_argument(dest="identifier", help="Table identifier")

    def run(self, input_args: Namespace):
        table_config = self._table_config_manager.get(input_args.identifier)

        try:
            self._table_deleter.delete(table_config)

            self._logger.info(f"Table {table_config.full_table_name} successfully deleted")
        except UnknownTableException as e:
            self._logger.error(str(e))
            sys.exit(1)



from datalakebundle.table.config import primitive_value
from datalakebundle.table.config.FieldsResolver import FieldsResolver
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.identifier.fill_template import fill_template
from datalakebundle.table.identifier.IdentifierParser import IdentifierParser


class TableConfigParser:
    def __init__(self, table_name_template: str):
        self.__fields_resolver = FieldsResolver()
        self.__identifier_parser = IdentifierParser()
        self.__table_name_template = table_name_template

    def parse(self, identifier: str, explicit_config: dict, defaults: dict = None):
        defaults = defaults or dict()
        identifiers = self.__identifier_parser.parse(identifier)
        table_name_parts = self.__resolve_table_name_parts(identifiers)

        if "partition_by" in explicit_config and isinstance(explicit_config["partition_by"], str):
            explicit_config["partition_by"] = [explicit_config["partition_by"]]

        all_fields = {**identifiers, **table_name_parts, **explicit_config}

        for name, val in self.__filter_primitive(all_fields, defaults).items():
            all_fields[name] = primitive_value.evaluate(val, all_fields)

        all_fields = self.__fields_resolver.resolve(all_fields, defaults)

        return TableConfig(**all_fields)

    def __filter_primitive(self, all_fields: dict, defaults: dict):
        return {name: val for name, val in defaults.items() if not isinstance(val, dict) and name not in all_fields}

    def __resolve_table_name_parts(self, identifiers: dict):
        full_table_name = fill_template(self.__table_name_template, identifiers)
        dot_position = full_table_name.find(".")

        if dot_position == -1:
            raise Exception("Table name must meet the following format: {dbName}.{tableName}")

        return {
            "db_name": full_table_name[:dot_position],
            "table_name": full_table_name[dot_position + 1 :],  # noqa: E203
        }



from databricksbundle.notebook.decorator.NodeDecorator import NodeDecorator


class DataFrameReturningDecorator(NodeDecorator):
    def _set_global_dataframe(self):
        self._function.__globals__[self._function.__name__ + "_df"] = self._result



from typing import Optional, List
from box import Box
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.config.TableConfigParser import TableConfigParser


class TableConfigManager:

    __table_configs: List[TableConfig] = []

    def __init__(
        self,
        raw_table_configs: Box,
        table_defaults: Box,
        table_config_parser: TableConfigParser,
    ):
        raw_table_configs = raw_table_configs or Box(dict())

        self.__table_configs = [
            table_config_parser.parse(identifier, explicit_config or dict(), table_defaults.to_dict())
            for identifier, explicit_config in raw_table_configs.to_dict().items()
        ]

    def exists(self, identifier: str) -> bool:
        for table_config in self.__table_configs:
            if table_config.identifier == identifier:
                return True

        return False

    def get(self, identifier: str) -> Optional[TableConfig]:
        for table_config in self.__table_configs:
            if table_config.identifier == identifier:
                return table_config

        raise Exception(f"Identifier {identifier} not found among datalakebundle.tables")

    def get_all(self):
        return self.__table_configs

    def get_by_filter(self, filter_function: callable):
        return list(filter(filter_function, self.__table_configs))

    def get_with_attribute(self, attr_name: str):
        return self.get_by_filter(lambda table_config: hasattr(table_config, attr_name))



import collections
from logging import Logger
from typing import Tuple
from pyspark.sql.dataframe import DataFrame
from datalakebundle.notebook.decorator.DataFrameReturningDecorator import DataFrameReturningDecorator


class DuplicateColumnsChecker:
    def __init__(
        self,
        logger: Logger,
    ):
        self.__logger = logger

    def check(self, df: DataFrame, result_decorators: Tuple[DataFrameReturningDecorator]):
        field_names = [field.name.lower() for field in df.schema.fields]
        duplicate_fields = dict()

        for field_name, count in collections.Counter(field_names).items():
            if count > 1:
                duplicate_fields[field_name] = []

        if duplicate_fields == dict():
            return

        fields2_tables = dict()

        for result_decorator in result_decorators:
            source_df = result_decorator.result
            for field in source_df.schema.fields:
                field_name = field.name.lower()

                if field_name not in fields2_tables:
                    fields2_tables[field_name] = []

                fields2_tables[field_name].append(result_decorator.function.__name__)

        for duplicate_field in duplicate_fields:
            self.__logger.error(f"Duplicate field {duplicate_field}", extra={"source_dataframes": fields2_tables[duplicate_field]})

        fields_string = ", ".join(duplicate_fields)
        raise Exception(f"Duplicate output column(s): {fields_string}. Disable by setting @transformation(check_duplicate_columns=False)")



from logging import Logger
from pyspark.sql.session import SparkSession
from pyspark.sql.dataframe import DataFrame
from datalakebundle.table.config.TableConfig import TableConfig
from datalakebundle.table.schema.DeltaTableSchema import DeltaTableSchema
import pyspark.sql.types as t
import yaml
import string
import random


class TableWriter:
    def __init__(
        self,
        logger: Logger,
        spark: SparkSession,
    ):
        self.__logger = logger
        self.__spark = spark

    def append(self, df: DataFrame, table_config: TableConfig):
        self.__insert_into(df, table_config, False)

    def overwrite(self, df: DataFrame, table_config: TableConfig):
        self.__insert_into(df, table_config, True)

    def upsert(self, df: DataFrame, table_config: TableConfig):
        if not isinstance(table_config.schema, DeltaTableSchema):
            raise Exception("In order to perform upsert operation schema must be of type DeltaTableSchema")

        if len(table_config.schema.primary_key_columns) == 0:
            raise Exception(f"Table {table_config.table_name} has no primary keys defined, define them in 'schema.py' file")

        self.__check_schema(df, table_config)

        temp_source_table = f"upsert_{table_config.table_identifier}_{''.join(random.choice(string.ascii_lowercase) for _ in range(6))}"

        df.createOrReplaceTempView(temp_source_table)

        upsert_sql_statement = self.__create_sql_upsert_statement(table_config, temp_source_table)

        try:
            self.__spark.sql(upsert_sql_statement)

        except BaseException:
            raise

        finally:
            self.__spark.catalog.dropTempView(temp_source_table)

    def write_if_not_exist(self, df: DataFrame, table_config: TableConfig):
        self.__check_schema(df, table_config)

        (
            df.write.partitionBy(table_config.partition_by)
            .format("delta")
            .option("overwriteSchema", "true")
            .mode("errorifexists")
            .saveAsTable(table_config.full_table_name, path=table_config.target_path)
        )

    def __check_schema(self, df: DataFrame, table_config: TableConfig):
        table_schema = table_config.schema

        def print_schema(schema: t.StructType):
            return yaml.dump(schema.jsonValue())

        def remove_metadata(json_schema):
            for field in json_schema["fields"]:
                field["metadata"] = dict()

            return json_schema

        if remove_metadata(table_schema.jsonValue()) != remove_metadata(df.schema.jsonValue()):
            self.__logger.warning(
                "Table and dataframe schemas do NOT match",
                extra={
                    "df_schema": print_schema(df.schema),
                    "table_schema": print_schema(table_schema),
                    "table_schema_loader": table_config.schema_loader,
                    "table": table_config.full_table_name,
                },
            )

    def __insert_into(self, df: DataFrame, table_config: TableConfig, overwrite: bool):
        self.__check_schema(df, table_config)

        df.write.insertInto(table_config.full_table_name, overwrite=overwrite)

    def __create_sql_upsert_statement(self, table_config: TableConfig, temp_source_table: str) -> str:
        conditions = []
        updates = []
        columns_to_update = set(table_config.schema.fieldNames()) - set(table_config.schema.primary_key_columns)

        for primary_key in table_config.schema.primary_key_columns:
            conditions.append(f"source.{primary_key} = target.{primary_key}")

        for col in columns_to_update:
            updates.append(f"target.{col} = source.{col}")

        statement = (
            f"MERGE INTO {table_config.full_table_name} AS target\n"
            f"USING {temp_source_table} AS source\n"
            f"ON {' AND '.join(conditions)}\n"
            f"{{matched_clause}}"
            f"WHEN NOT MATCHED THEN INSERT *\n"
        )

        if len(updates) > 0:
            statement = statement.format(matched_clause=f"WHEN MATCHED THEN UPDATE SET {', '.join(updates)}\n")
        else:
            statement = statement.format(matched_clause="")

        return statement
